@@ -8,11 +8,12 @@ my %sources;
 sub eval_in_ctx($ctx, $code) {
     ENTER $*DEBUG_HOOKS.suspend();
     LEAVE $*DEBUG_HOOKS.unsuspend();
-    my $compiler := pir::compreg__PS('perl6');
-    my $vm_ctx   := nqp::getattr(nqp::p6decont($ctx), PseudoStash, '$!ctx');
-    my $pbc      := $compiler.compile($code, :outer_ctx($vm_ctx), :global(GLOBAL));
-    nqp::atpos($pbc, 0).set_outer_ctx($vm_ctx);
-    $pbc();
+    my $compiler := nqp::getcomp('perl6');
+    my $vm_ctx   := nqp::getattr(nqp::decont($ctx), PseudoStash, '$!ctx');
+    my $comp'd   := nqp::findmethod($compiler, 'compile')($compiler,
+                        $code, :outer_ctx($vm_ctx), :global(GLOBAL));
+    nqp::forceouterctx($comp'd, $vm_ctx);
+    $comp'd();
 }
 
 # Represents a file that we're debugging.
@@ -22,6 +23,12 @@ my class SourceFile {
     has @!lines;
     has @!line_offsets;
     has @!regex_regions;
+    has %!routine_regions{Range};
+    
+    my class RoutineInfo {
+        has $.type;
+        has $.name;
+    }
     
     method BUILD(:$!filename, :$!source) {
         # Ensure source ends with a newline.
@@ -45,6 +52,19 @@ my class SourceFile {
         @!regex_regions.push(item $from_pos..$to_pos);
     }
     
+    method add_routine_region($from_pos, $to_pos, $type, $name) {
+        %!routine_regions{item $from_pos..$to_pos} = RoutineInfo.new(:$type, :$name);
+    }
+    
+    method routine_containing($from_pos, $to_pos) {
+        my @best = %!routine_regions.grep(
+                { $from_pos..$to_pos ~~ $^r.key }
+            ).sort(
+                { .key.max - .key.min }
+            );
+        @best ?? @best[0].value.name !! ''
+    }
+    
     method line_of($pos, $def_line, $def_pos) {
         my $last_p = 0;
         for @!line_offsets.kv -> $l, $p {
@@ -176,14 +196,53 @@ my class SourceFile {
     }
 }
 
+# File and line position.
+my regex flpos {
+    [$<file>=[<-[:]>+] ':']? $<line>=[\d+]
+}
+
 # Holds the current state of the debugger.
 my class DebugState {
-    my enum RunMode <Step RunToThrowOrBreakpoint RunToUnhandledOrBreakpoint>;
+    my enum RunMode <Step StepOver StepOut RunToThrowOrBreakpoint RunToUnhandledOrBreakpoint>;
     my RunMode $run_mode  = Step;
     my Bool    $dying     = False;
     my Bool    $in_prompt = False;
-    my %breakpoints;
     my $cur_ex;
+    my %stepping_out_of;
+    my %stepping_over_in;
+    my %breakpoints;
+    my %tracepoints;
+    my @tp_log;
+    
+    my class TracePointLogEntry {
+        has $.tp;
+        has $.result;
+        has $.fail;
+    }
+    
+    my class TracePoint {
+        has $.file;
+        has $.line;
+        has $.expr;
+        
+        method hit($ctx) {
+            try {
+                @tp_log.push(TracePointLogEntry.new(
+                    tp     => self,
+                    result => eval_in_ctx($ctx, $!expr).gist
+                ));
+                CATCH {
+                    default {
+                        @tp_log.push(TracePointLogEntry.new(
+                            tp     => self,
+                            result => .gist,
+                            fail   => True
+                        ));
+                    }
+                }
+            }
+        }
+    }
 
     method set_current_exception($ex) {
         $cur_ex = $ex;
@@ -246,8 +305,82 @@ my class DebugState {
         }
     }
     
+    method add_tracepoint($file, $line, $expr) {
+        if self.normalize_filename($file) -> $norm_file {
+            push %tracepoints{$norm_file}, TracePoint.new(:file($norm_file), :$line, :$expr);
+        }
+        else {
+            say colored("Cannot add tracepoint to unknown file '$file'", 'red');
+        }
+    }
+    
+    method log_tracepoints($ctx, $filename, $from, $to) {
+        if %tracepoints{$filename} -> @tps {
+            my ($from_line, $) = %sources{$filename}.line_of($from, -1, -1);
+            my ($to_line, $)   = %sources{$filename}.line_of($to, -1, -1);
+            @tps.grep({ $^tp.line - 1 ~~ $from_line..$to_line })>>.hit($ctx);
+        }
+    }
+    
+    method render_one_tracepoint($tp) {
+        say colored(">>> ", 'blue') ~ $tp.expr;
+        for @tp_log.grep(*.tp === $tp) {
+            say .fail
+                ?? colored("* $_.result()", 'red')
+                !! colored("* ", 'blue') ~ $_.result();
+        }
+    }
+    
+    method render_all_tracepoints() {
+        my $last_tp;
+        for @tp_log -> (:$tp, :$result, :$fail) {
+            unless $tp === $last_tp {
+                say colored(">>> ", 'blue') ~ "$tp.file():$tp.line()";
+            }
+            say $fail
+                ?? colored("* $result", 'red')
+                !! colored("* ", 'blue') ~ $result;
+            $last_tp = $tp;
+        }
+    }
+    
     method should_break_at($filename, $from, $to) {
-        $run_mode == Step || self.is_breakpoint_at($filename, $from, $to)
+        given $run_mode {
+            when Step {
+                True
+            }
+            when StepOver {
+                ENTER $in_prompt = True;
+                LEAVE $in_prompt = False;
+                my $depth = +lines(Backtrace.new().full) - 1;
+                if $filename eq %stepping_over_in<file> &&
+                        %sources{$filename}.routine_containing($from, $to) eq %stepping_over_in<routine> &&
+                        $depth == %stepping_over_in<depth>
+                        || $depth < %stepping_over_in<depth> {
+                    $run_mode = Step;
+                    %stepping_over_in = ();
+                    True
+                }
+                else {
+                    False
+                }
+            }
+            when StepOut {
+                if ($filename ne %stepping_out_of<file> || 
+                         %sources{$filename}.routine_containing($from, $to) ne %stepping_out_of<routine>)
+                         && (+lines(Backtrace.new().full) - 1) < %stepping_out_of<depth> {
+                    $run_mode = Step;
+                    %stepping_out_of = ();
+                    True
+                }
+                else {
+                    False
+                }
+            }
+            default {
+                self.is_breakpoint_at($filename, $from, $to);
+            }
+        }
     }
     
     method should_break_on_throw() {
@@ -275,8 +408,8 @@ my class DebugState {
             take $c unless $c eq "\r";
         }
     }
-    
-    method issue_prompt($ctx, $cur_file) {
+
+    method issue_prompt($ctx, $cur_file, $from = 0, $to = 0) {
         ENTER $in_prompt = True;
         LEAVE $in_prompt = False;
         loop {
@@ -306,7 +439,7 @@ my class DebugState {
                         }
                     }
                 }
-                when /^ (< $ @ % > .+)/ {
+                when /^ (< $ @ % > .+ | 'self' .*)/ {
                     say eval_in_ctx($ctx, ~$0).perl;
                     CATCH {
                         default {
@@ -332,6 +465,38 @@ my class DebugState {
                         return;
                     }
                 }
+                when 's' {
+                    if $dying {
+                        self.complain_about_being_dying();
+                    }
+                    else {
+                        my $cur_routine = %sources{$cur_file}.routine_containing($from, $to);
+                        $run_mode = StepOver;
+                        %stepping_over_in =
+                            file    => $cur_file,
+                            routine => $cur_routine,
+                            depth   => +lines(Backtrace.new().full) - 4;
+                        return;
+                    }
+                }
+                when 'so' {
+                    if $dying {
+                        self.complain_about_being_dying();
+                    }
+                    else {
+                        if %sources{$cur_file}.routine_containing($from, $to) -> $cur_routine {
+                            $run_mode = StepOut;
+                            %stepping_out_of =
+                                file    => $cur_file,
+                                routine => $cur_routine,
+                                depth   => +lines(Backtrace.new().full) - 4;
+                            return;
+                        }
+                        else {
+                            say colored("Not currently in a routine; cannot step out", 'red');
+                        }
+                    }
+                }
                 when 'bt' | 'st' {
                     say join "\n", lines(Backtrace.new().nice)[4..*];
                 }
@@ -343,8 +508,8 @@ my class DebugState {
                         say colored('No current exception', 'red');
                     }
                 }
-                when /^ 'bp' <.ws> 'add' <.ws> [$<file>=[<-[:]>+] ':']? $<line>=[\d+] $/ {
-                    self.add_breakpoint($<file> ?? ~$<file> !! $cur_file, +$<line>);
+                when /^ 'bp' <.ws> 'add' <.ws> <p=&flpos> $/ {
+                    self.add_breakpoint($<p><file> ?? ~$<p><file> !! $cur_file, +$<p><line>);
                 }
                 when /^ 'bp' <.ws> 'list' $/ {
                     for %breakpoints.kv -> $file, @lines {
@@ -354,12 +519,44 @@ my class DebugState {
                         }
                     }
                 }
-                when /^ 'bp' <.ws> 'rm' <.ws> [$<file>=[<-[:]>+] ':']? $<line>=[\d+] $/ {
-                    self.remove_breakpoint($<file> ?? ~$<file> !! $cur_file, +$<line>);
+                when /^ 'bp' <.ws> 'rm' <.ws> <p=&flpos> $/ {
+                    self.remove_breakpoint($<p><file> ?? ~$<p><file> !! $cur_file, +$<p><line>);
                 }
                 when /^ 'bp' <.ws> 'rm' <.ws> 'all' $/ {
                     %breakpoints = ();
                 }
+                when /^ 'tp' <.ws> 'add' <.ws> <p=&flpos> <.ws> (.+) $/ {
+                    self.add_tracepoint($<p><file> ?? ~$<p><file> !! $cur_file, +$<p><line>, ~$0);
+                }
+                when /^ 'tp' <.ws> 'list' $/ {
+                    for %tracepoints.kv -> $file, @tps {
+                        if @tps {
+                            say "$file:";
+                            for @tps -> $tp {
+                                say "    $tp.line(): $tp.expr()";
+                            }
+                        }
+                    }
+                }
+                when /^ 'tp' <.ws> 'show' [ <.ws> <p=&flpos> ]? $/ {
+                    if $<p> {
+                        my $file = $<p>[0]<file> ?? ~$<p>[0]<file> !! $cur_file;
+                        if %tracepoints{self.normalize_filename($file)} -> @tps {
+                            if @tps.first(*.line == +$<p>[0]<line>) -> $tp {
+                                self.render_one_tracepoint($tp);
+                            }
+                            else {
+                                say colored('No tracepoint at this line', 'red');
+                            }
+                        }
+                        else {
+                            say colored('No tracepoint in this file', 'red');
+                        }
+                    }
+                    else {
+                        self.render_all_tracepoints();
+                    }
+                }
                 when '?' | 'h' | 'help' {
                     say self.usage()
                 }
@@ -380,23 +577,39 @@ my class DebugState {
     
     method usage() {
         join "\n",
-            ('<enter>            single step' unless $dying),
-            ('r                  run until the next breakpoint or unhnadled exception' unless $dying),
-            ('rt                 run until the next breakpoint or an exception is thrown' unless $dying),
-            's[ay], p[rint]     evaluate and display an expression in the current scope',
-            'e[val]             evaluate an expression in the current scope',
-            '$s, @a, %h         show .perl of the a variable in scope (indexing allowed)',
-            'bt, st             show the backtrace from the current location',
-            ('ex                 show .perl of the current exception' if $cur_ex),
-            'bp add file:line   adds a breakpoint at the specified file/line',
-            'bp add line        adds a breakpoint at the specified line in this file',
-            'bp list            lists all active breakpoints',
-            'bp rm file:line    removes the breakpoint at the specified file/line',
-            'bp rm line         removes the breakpoint at the specified line in this file',
-            'bp rm all          removes all breakpoints',
-            'q[uit]             exit the debugger'
+            ('<enter>                single step, stepping into any calls' unless $dying),
+            ('s                      step to next statement, stepping over any calls' unless $dying),
+            ('so                     step out of the current routine' unless $dying),
+            ('r                      run until the next breakpoint or unhnadled exception' unless $dying),
+            ('rt                     run until the next breakpoint or an exception is thrown' unless $dying),
+            's[ay], p[rint]         evaluate and display an expression in the current scope',
+            'e[val]                 evaluate an expression in the current scope',
+            '$s, @a, %h, self       show .perl of the a variable in scope (indexing allowed)',
+            'bt, st                 show the backtrace from the current location',
+            ('ex                     show .perl of the current exception' if $cur_ex),
+            'bp add file:line       adds a breakpoint at file/line (file optional)',
+            'bp list                lists all active breakpoints',
+            'bp rm file:line        removes the breakpoint at file/line (file optional)',
+            'bp rm all              removes all breakpoints',
+            'tp add file:line expr  logs the value of expr each time file:line is hit',
+            'tp list                shows a list of all set tracepoints',
+            'tp show                shows the data collected by all tracepoints',
+            'tp show file:line      shows the data collected by the specified tracepoint',
+            'q[uit]                 exit the debugger'
             ;
     }
+    
+    method reset() {
+        $run_mode  = Step;
+        $dying     = False;
+        $in_prompt = False;
+        $cur_ex = Mu;
+        %stepping_out_of = ();
+        %stepping_over_in = ();
+        %breakpoints = ();
+        %tracepoints = ();
+        @tp_log = ();
+    }
 }
 
 # Install various hooks.
@@ -406,27 +619,36 @@ $*DEBUG_HOOKS.set_hook('new_file', -> $filename, $source {
     }
     %sources{$filename} = SourceFile.new(:$filename, :$source);
 });
+$*DEBUG_HOOKS.set_hook('routine_region', -> $filename, $from_pos, $to_pos, $type, $name {
+    %sources{$filename}.add_routine_region($from_pos, $to_pos, $type, $name);
+});
 $*DEBUG_HOOKS.set_hook('statement_simple', -> $filename, $ctx, $from, $to {
+    DebugState.log_tracepoints($ctx, $filename, $from, $to);
     if DebugState.should_break_at($filename, $from, $to) {
         say %sources{$filename}.summary_around($from, $to, $ctx);
-        DebugState.issue_prompt($ctx, $filename);
+        DebugState.issue_prompt($ctx, $filename, $from, $to);
     }
 });
 $*DEBUG_HOOKS.set_hook('statement_cond', -> $filename, $ctx, $type, $from, $to {
+    DebugState.log_tracepoints($ctx, $filename, $from, $to);
     if DebugState.should_break_at($filename, $from, $to) {
         say %sources{$filename}.summary_around($from, $to, $ctx);
-        DebugState.issue_prompt($ctx, $filename);
+        DebugState.issue_prompt($ctx, $filename, $from, $to);
     }
 });
 $*DEBUG_HOOKS.set_hook('regex_region', -> $filename, $from_pos, $to_pos {
     %sources{$filename}.add_regex_region($from_pos, $to_pos);
 });
 $*DEBUG_HOOKS.set_hook('regex_atom', -> $filename, $ctx, $from, $to {
+    DebugState.log_tracepoints($ctx, $filename, $from, $to);
     if DebugState.should_break_at($filename, $from, $to) {
         say %sources{$filename}.summary_around($from, $to, $ctx);
-        DebugState.issue_prompt($ctx, $filename);
+        DebugState.issue_prompt($ctx, $filename, $from, $to);
     }
 });
+$*DEBUG_HOOKS.set_hook('reset', -> {
+    DebugState.reset();
+});
 
 # Allow interception of throwing an exception.
 my $IN_UNHANDLED = 0;
@@ -450,7 +672,12 @@ sub thrown(|) {
     my $bt = $e.backtrace();
     my $ctx = CALLER;
     my ($file, $line);
+    my $fail = False;
     for @$bt {
+        if .name eq '&fail' {
+            $fail = True;
+            last;
+        }
         if %sources.exists(.file) {
             $file = .file;
             $line = .line;
@@ -458,7 +685,7 @@ sub thrown(|) {
         }
         $ctx = $ctx.WHO.<CALLER>;
     }
-    if $file {
+    if !$fail && $file {
         DebugState.set_current_exception($e);
         say %sources{$file}.throw_summary($e, $line - 1);
         DebugState.issue_prompt($ctx.WHO, $file);
@@ -466,7 +693,7 @@ sub thrown(|) {
 }
 
 # Override handler for uncaught exceptions.
-my Mu $p6comp := pir::compreg__Ps('perl6');
+my Mu $p6comp := nqp::getcomp('perl6');
 $p6comp.HOW.find_method($p6comp, 'handle-exception').wrap(-> | {
     my Mu $vm_ex := nqp::atpos(pir::perl6_current_args_rpa__P(), 1);
     pir::perl6_invoke_catchhandler__vPP(&unhandled, $vm_ex);
