@@ -29,19 +29,19 @@ multi to-json(Str:D  $d) {
             .subst(/<-[\c32..\c126]>/, { ord(~$_).fmt('\u%04x') }, :g)
     ~ '"'
 }
-multi to-json(Array:D $d) {
+multi to-json(Positional:D $d) {
     return  '[ '
             ~ $d.map(&to-json).join(', ')
             ~ ' ]';
 }
-multi to-json(Hash:D  $d) {
+multi to-json(Associative:D  $d) {
     return '{ '
             ~ $d.map({ to-json(.key) ~ ' : ' ~ to-json(.value) }).join(', ')
             ~ ' }';
 }
 
-multi to-json(Any:U $) { 'null' }
-multi to-json(Any:D $s) {
+multi to-json(Mu:U $) { 'null' }
+multi to-json(Mu:D $s) {
     die "Can't serialize an object of type " ~ $s.WHAT.perl
 }
 
